import numpy as np
from nn.activations.activation import Activation

class Tanh(Activation):
    def __init__(self):
        def tanh(x):
            return np.tanh(x)

        def tanh_prime(x):
            return 1 - np.tanh(x) ** 2

        super().__init__(tanh, tanh_prime)

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
# Solves a niche error when the input to the log is zero.
from scipy.special import log1p

# https://neuralthreads.medium.com/binary-cross-entropy-loss-special-case-of-categorical-cross-entropy-loss-95c0c338d183
def binary_cross_entropy(y_true, y_pred):
    return np.mean(-y_true * log1p(y_pred) - (1 - y_true) * log1p(1 - y_pred))

def binary_cross_entropy_prime(y_true, y_pred):
    return ((1 - y_true) / (1 - y_pred) - y_true / y_pred) / np.size(y_true)


import numpy as np
from nn.layers.layer import Layer

class Reshape(Layer):
    def __init__(self, input_shape, output_shape):
        self.input_shape = input_shape
        self.output_shape = output_shape

    def forward(self, input):
        return np.reshape(input, self.output_shape)

    def backward(self, output_gradient):
        return np.reshape(output_gradient, self.input_shape)

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "({}, {})".format(self.input_shape, self.output_shape)


import numpy as np

# https://neuralthreads.medium.com/categorical-cross-entropy-loss-the-most-important-loss-function-d3792151d05b
def categorical_cross_entropy(y_true, y_pred):
    return -np.sum(y_true * np.log(y_pred + 10**-100))

def categorical_cross_entropy_prime(y_true, y_pred):
    return -y_true/(y_pred + 10**-100)


import numpy as np
from nn.activations.activation import Activation

class LeakyRelu(Activation):
    def __init__(self):
        def leaky_relu(x):
            alpha = 0.01
            return np.maximum(alpha * x, x)

        def leaky_relu_prime(x):
            alpha = 0.01
            dx = np.ones_like(x)
            dx[x < 0] = alpha
            return dx

        super().__init__(leaky_relu, leaky_relu_prime)
    
    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
from nn.layers.layer import Layer

class Softmax(Layer):
    def forward(self, input):
        # More stable softmax
        #https://stackoverflow.com/questions/54880369/implementation-of-softmax-function-returns-nan-for-high-inputs
        tmp = np.exp(input - max(input))
        self.output = tmp / np.sum(tmp)
        return self.output
    
    # Might need to re-evaluate this backprop after making the forward stable softmax
    def backward(self, output_gradient):
        n = np.size(self.output)
        return np.dot((np.identity(n) - self.output.T) * self.output, output_gradient)
    
    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
from nn.activations.activation import Activation

class Relu(Activation):
    def __init__(self):
        def relu(x):
            return np.maximum(0, x)

        def relu_prime(x):
            return np.where(x >= 0, 1, 0)

        super().__init__(relu, relu_prime)
    
    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
from nn.layers.layer import Layer

class Activation(Layer):
    def __init__(self, activation, activation_prime):
        self.activation = activation
        self.activation_prime = activation_prime

    def forward(self, input):
        self.input = input
        return self.activation(self.input)

    def backward(self, output_gradient):
        return np.multiply(output_gradient, self.activation_prime(self.input))
    
    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
from scipy import signal
from nn.layers.layer import Layer
from nn.layers.layer_properties import LayerProperties
from nn.initializers import *
from nn.optimizers import *
from copy import deepcopy
from typing import Tuple

class Convolutional(Layer):

    def __init__(self, input_shape, kernel_size, kernel_depth: int, stride: Tuple[int,int]= (1,1), padding = (0), bias_mode: str = 'untied', layer_properties: LayerProperties = None):
        # Default layer properties
        self.layer_properties = LayerProperties(learning_rate=0.05, weight_initializer=Uniform(), bias_initializer=Uniform(), optimizer=SGD())

        # Parse and update any optional layer properties
        self.parse_layer_properties(layer_properties)

        self.kernel_depth = kernel_depth
        self.input_shape = input_shape
        self.input_depth = input_shape[0]
        self.kernel_size = self.parse_kernel(kernel_size)
        self.kernels_shape = (self.kernel_depth, self.input_depth, self.kernel_size[0], self.kernel_size[1])
        self.kernels = self.layer_properties.weight_initializer.get(*self.kernels_shape)

        self.stride = stride
        # Parse the padding into a 3D tuple ((padding_height1, padding_width1), (padding_height2, padding_width2), (padding_height2, padding_width2))
        # One tuple pair for each dimension. (depth, kernel height, kernel width)
        self.padding = self.parse_padding(padding)

        self.output_shape = (self.kernel_depth,
                            (self.input_shape[1] - self.kernel_size[0] + (self.padding[1][0] + self.padding[1][1])) // self.stride[0] + 1,
                            (self.input_shape[2] - self.kernel_size[1] + (self.padding[2][0] + self.padding[2][1])) // self.stride[1] + 1)

        valid_bias_modes = {'tied', 'untied'}
        if bias_mode not in valid_bias_modes:
            raise ValueError("bias_mode must be one of {}, but got bias_mode='{}'".format(valid_bias_modes, bias_mode))
        self.bias_mode = bias_mode

        if bias_mode == 'untied':
            self.biases = self.layer_properties.bias_initializer.get(*self.output_shape)
        elif bias_mode == 'tied':
            self.biases = self.layer_properties.bias_initializer.get(self.kernel_depth, 1, 1)
        
        # Run sanity checks to eliminate input errors
        self.sanity_checks()

    def forward(self, input):
        if self.bias_mode == 'untied':
            return self.forward_untied(input)
        elif self.bias_mode == 'tied':
            return self.forward_tied(input)     
    
    def backward(self, output_gradient):
        if self.bias_mode == 'untied':
            return self.backward_untied(output_gradient)
        elif self.bias_mode == 'tied':
            return self.backward_tied(output_gradient)

    # Separated Untied and Tied implementations into their own functions
    # Untied bias: where you use use one bias per kernel and output
    # Tied bias: where you share one bias per kernel
    def forward_untied(self, input):
        input = np.pad(input, self.padding, mode='constant')
        self.input = input
        self.output = np.copy(self.biases)
        for i in range(self.kernel_depth):
            for j in range(self.input_depth):
                self.output[i] += signal.correlate2d(input[j], self.kernels[i, j], "valid")[::self.stride[0], ::self.stride[1]]
        return self.output

    def backward_untied(self, output_gradient):
        # https://medium.com/@mayank.utexas/backpropagation-for-convolution-with-strides-8137e4fc2710
        # link to the dilation concept is above
        # The output gradient must be dilated before it can be correlated with the input due to striding.
        # The 1 * min is basically just a toggle. Because in the case of a stride of 1 you do NOT want to subtract 1 or dilate the output.
        depth, height, width = output_gradient.shape
        dilated_output_gradient = np.zeros((depth,
                        height + ((self.stride[0] - 1) * height - (1 * min(self.stride[0] - 1, 1))),
                        width + ((self.stride[1] - 1) * width - (1 * min(self.stride[1] - 1, 1)))),
                        dtype=output_gradient.dtype)
        dilated_output_gradient[::, ::self.stride[0], ::self.stride[1]] = output_gradient

        # If there is a non zero padding of we need to crop the output_gradient to match the input gradient.
        # This results in minor data loss but is easier to implement.
        # [top:bottom, left:right]
        dilated_trunc_output_gradient = dilated_output_gradient
        if self.padding[1][0] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, self.padding[1][0]:, :]
        if self.padding[1][1] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :-self.padding[1][1], :]
        if self.padding[2][0] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :, self.padding[2][0]:]
        if self.padding[2][1] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :, :-self.padding[2][1]]

        kernels_gradient = np.zeros(self.kernels_shape)
        input_gradient = np.zeros(self.input_shape)
        for i in range(self.kernel_depth):
            for j in range(self.input_depth):
                kernels_gradient[i, j] = signal.correlate2d(self.input[j], dilated_output_gradient[i], "valid")
                input_gradient[j] += signal.convolve2d(dilated_trunc_output_gradient[i], self.kernels[i, j], "full")

        self.kernels += self.layer_properties.weight_optimizer.calc(self.layer_properties.learning_rate, kernels_gradient)
        self.biases += self.layer_properties.bias_optimizer.calc(self.layer_properties.learning_rate, output_gradient)
        return input_gradient
    
    def forward_tied(self, input):
        input = np.pad(input, self.padding, mode='constant')
        self.input = input
        self.output = np.zeros(self.output_shape)
        for i in range(self.kernel_depth):
            for j in range(self.input_depth):
                self.output[i] += signal.correlate2d(self.input[j], self.kernels[i, j], "valid")[::self.stride[0], ::self.stride[1]]
            self.output[i] += self.biases[i]
        return self.output

    def backward_tied(self, output_gradient):
        depth, height, width = output_gradient.shape
        dilated_output_gradient = np.zeros((depth,
                        height + ((self.stride[0] - 1) * height - (1 * min(self.stride[0] - 1, 1))),
                        width + ((self.stride[1] - 1) * width - (1 * min(self.stride[1] - 1, 1)))),
                        dtype=output_gradient.dtype)
        dilated_output_gradient[::, ::self.stride[0], ::self.stride[1]] = output_gradient

        # If there is a non zero padding of we need to crop the output_gradient
        # [top:bottom, left:right]
        dilated_trunc_output_gradient = dilated_output_gradient
        if self.padding[1][0] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, self.padding[1][0]:, :]
        if self.padding[1][1] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :-self.padding[1][1], :]
        if self.padding[2][0] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :, self.padding[2][0]:]
        if self.padding[2][1] != 0:
            dilated_trunc_output_gradient = dilated_trunc_output_gradient[:, :, :-self.padding[2][1]]

        kernels_gradient = np.zeros(self.kernels_shape)
        input_gradient = np.zeros(self.input_shape)
        for i in range(self.kernel_depth):
            for j in range(self.input_depth):
                kernels_gradient[i, j] = signal.correlate2d(self.input[j], dilated_output_gradient[i], "valid")
                input_gradient[j] += signal.convolve2d(dilated_trunc_output_gradient[i], self.kernels[i, j], "full")
        self.kernels += self.layer_properties.weight_optimizer.calc(self.layer_properties.learning_rate, kernels_gradient)
        # Sum all the gradients for the output gradient of each kernel then multiply by the learning rate.
        self.biases += self.layer_properties.bias_optimizer.calc(self.layer_properties.learning_rate, np.sum(output_gradient, keepdims=True, axis=(1,2)))
        return input_gradient
    
    def parse_kernel(self, kernel):
        # checks for single integer
        if isinstance(kernel, int):
            return (kernel, kernel)
        # checks for single value in tuple
        elif isinstance(kernel, tuple) and len(kernel) == 1:
            return (kernel[0], kernel[0])
        elif isinstance(kernel, tuple) and len(kernel) == 2:
            # No changes necessary if tuple with 2 values.
            return kernel
        else:
            raise ValueError("Kernel size invalid. Please use one of the following formats...\n" \
                            "kernel_size = 3\n" \
                            "kernel_size = (3)\n" \
                            "kernel_size = (3,3)\n")
    
    def parse_padding(self, padding):
        # checks for single integer
        if isinstance(padding, int):
            # The extra tuple of zeros is for the depth dimension which needs no padding.
            return ((0, 0), (padding, padding), (padding, padding))
        # checks for single value in tuple
        elif isinstance(padding, tuple) and len(padding) == 1:
            return ((0, 0), (padding[0], padding[0]), (padding[0], padding[0]))
        elif isinstance(padding, tuple) and len(padding) == 2:
            # (height, width) or (3, 3)
            if isinstance(padding[0], int) and isinstance(padding[1], int):
                return ((0, 0), (padding[0], padding[0]), (padding[1], padding[1]))
            # ((top, bottom), (left, right))
            elif isinstance(padding[0], tuple) and isinstance(padding[1], tuple) and len(padding[0]) == 2 and len(padding[1]) == 2:
                return ((0, 0), (padding[0][0], padding[0][1]), (padding[1][0], padding[1][1]))
        else:
            raise ValueError("Padding size invalid. Please use one of the following formats...\n" \
                            "padding = 3\n" \
                            "padding = (3)\n" \
                            "padding = (3,3)\n" \
                            "padding = ((3,3), (3,3))\n")
    
    def parse_layer_properties(self, layer_properties):
        # Optionally set the layer properties for all layers that utilize layer properties parameters
        if layer_properties is not None:
            # Replace all layer defaults with any non "None" layer properties.
            # This is just a lot of fancy code to allow you to override only 'some' of the default layer properties.
            # Instead of forcing you to populate all the parameters every time.
            for attr, _ in layer_properties.__dict__.items():
                if getattr(layer_properties, attr) is not None:
                    # copy is necessary to ensure that individual layer classes don't get shared instances of an optimizer
                    # optimizers such as momentum sgd require separate instances to track velocity
                    setattr(self.layer_properties, attr, deepcopy(getattr(layer_properties, attr)))

    def sanity_checks(self):
        # Error checking to make sure the kernel and stride can evenly map across the input.
        dimension_1 = (self.input_shape[1] - self.kernel_size[0] + (self.padding[1][0] + self.padding[1][1])) / self.stride[0] + 1
        dimension_2 = (self.input_shape[2] - self.kernel_size[1] + (self.padding[2][0] + self.padding[2][1])) / self.stride[1] + 1
        if not dimension_1.is_integer() or not dimension_2.is_integer():
            raise ValueError("(input_size - kernel_size + (2 * padding)) / stride ... must be an integer.\n" \
                            "dimension 0 = depth\n" \
                            "dimension 1 = input height\n" \
                            "dimension 2 = input width\n" \
                            "This library does not handle cropping inputs for kernels / strides that do not evenly divide the input.")
        
        # If the padding is larger than the kernel size then you have values that will always be zero.
        # Because the output will only be calculated based on padding and not any input data
        if self.padding[1][0] >= self.kernel_size[0]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[1][0], self.kernel_size))
        elif self.padding[1][1] >= self.kernel_size[0]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[1][1], self.kernel_size))
        elif self.padding[2][0] >= self.kernel_size[1]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[2][0], self.kernel_size))
        elif self.padding[2][1] >= self.kernel_size[1]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[2][1], self.kernel_size))

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "(({}, {}, {}), kernel_size = {}, kernel_depth = {}, bias_mode = {}, stride = {}, padding = {})".format(
            self.input_shape[0],
            self.input_shape[1],
            self.input_shape[2],
            self.kernel_size,
            self.kernel_depth,
            self.bias_mode,
            self.stride,
            self.padding[1:]
        )


class TrainingSet():

    def __init__(self, input_train, output_train, input_test, output_test, post_processing = lambda x : x):
        # Datasets (numpy arrays)
        # Training Set
        self._input_train = input_train
        self._output_train = output_train
        # Validation Set
        self._input_test = input_test
        self._output_test = output_test

        # Applied to the network prediction and training data before equivalency comparison.
        # By default it applies no post processing
        self._post_processing = post_processing
        
        # Calculate dataset sizes
        self._input_train_size = len(self.input_train)
        self._output_train_size = len(self.output_train)
        self._input_test_size = len(self.input_test)
        self._output_test_size = len(self.output_test)
    
    @property
    def input_train(self):
        return self._input_train
    
    @property
    def output_train(self):
        return self._output_train

    @property
    def input_test(self):
        return self._input_test
    
    @property
    def output_test(self):
        return self._output_test
    
    @property
    def post_processing(self):
        return self._post_processing

    @property
    def input_train_size(self):
        return self._input_train_size
    
    @property
    def output_train_size(self):
        return self._output_train_size

    @property
    def input_test_size(self):
        return self._input_test_size
    
    @property
    def output_test_size(self):
        return self._output_test_size

    def __str__(self):
        training_str = "{:<15} {} {}".format("Training Data:", self.input_train_size, "samples")
        test_str = "{:<15} {} {}".format("Test Data:", self.input_test_size, "samples")
        return training_str + '\n' + test_str


import numpy as np
from nn.layers.layer import Layer
from nn.initializers import *
from nn.optimizers import *
from typing import Tuple
from numpy.lib.stride_tricks import as_strided

# Helpful reference articles
# https://stackoverflow.com/questions/54962004/implement-max-mean-poolingwith-stride-with-numpy
# https://towardsdatascience.com/advanced-numpy-master-stride-tricks-with-25-illustrated-exercises-923a9393ab20#1f28
# https://deepnotes.io/maxpool

# In the future I would ideally like to use the numpy array directly to map the output gradient indices.
# https://stackoverflow.com/questions/26194389/how-to-rearrange-array-based-upon-index-array
# I feel like it would probably make this class more efficient

# The max pooling layer does not change the depth of the input. And it has no trainable parameters.
class MaxPooling2D(Layer):

    def __init__(self, input_shape, kernel_size, stride: Tuple[int,int]= (2,2), padding = (0)):
        self.input_shape = input_shape
        self.kernel_size = self.parse_kernel(kernel_size)
        self.stride = stride
        self.padding = self.parse_padding(padding)
        
        # The final output shape of the MaxPooling layer
        self.output_shape = (self.input_shape[0], 
                            (self.input_shape[1] - self.kernel_size[0] + (self.padding[1][0] + self.padding[1][1])) // self.stride[0] + 1,
                            (self.input_shape[2] - self.kernel_size[1] + (self.padding[2][0] + self.padding[2][1])) // self.stride[1] + 1)
        # This index map contains the input indexes that the output gradient maps to.
        # Basically the value at index_map[depth][x][y] is coordinates to the input that output_gradient[depth][x][y] needs to be applied to.
        self.index_map_x = np.zeros((self.output_shape[0], self.output_shape[1], self.output_shape[2]), dtype="int")
        self.index_map_y = np.zeros((self.output_shape[0], self.output_shape[1], self.output_shape[2]), dtype="int")

        # Run sanity checks to eliminate input errors
        self.sanity_checks()

    def forward(self, input):
        # Padding. Have to pad with negative infinity for max pooling.
        # Otherwise padding could be identified as the "max" if the inputs are negative values.
        input = np.pad(input, self.padding, mode='constant', constant_values=np.NINF)
        
        # The array is 5D because it correlates each stride calculated with each output. (depth, output_x, output_y, stride_result_x, stride_result_y)
        # This is a pre-processing step so we can call the max function on each stride output and map them to the output shape.
        view_shape = (self.output_shape[0], self.output_shape[1], self.output_shape[2], self.kernel_size[0], self.kernel_size[1])
        # The strides that will be applied to the input array
        view_strides = (input.strides[0], self.stride[0] * input.strides[1], self.stride[1] * input.strides[2], input.strides[1], input.strides[2])
        
        # 5D array containing the depth and all the striding 2D outputs mapped to the 2D pooling output space.
        strided_result = as_strided(input, view_shape, view_strides)
                
        # Result of max pooling along the stride axes
        out = strided_result.max(axis=(3, 4))

        # Flatten the last two array dimensions because argmax only supports flat arrays. Now a 4D array.
        argmax_arr = np.reshape(strided_result, (strided_result.shape[0], strided_result.shape[1], strided_result.shape[2], strided_result.shape[3] * strided_result.shape[4]))
        # Find the index of the max value on the flat array (axis 3). argmax only takes the index of the first max value if there are duplicates.
        argmax_arr = np.argmax(argmax_arr, axis=3)

        # Calculate the input index based on the stride's local index, the strided_result array index, and the stride itself.
        # Then store that input index to be used as a sort of reordering / mask for the output gradient in backprop.
        # self.index_map_x[depth][x][y] = local_stride_index_x + (strided_result_x_index * self.stride[0])
        # self.index_map_y[depth][x][y] = local_stride_index_y + (strided_result_y_index * self.stride[1])

        # indices gives me a mask for the stride's coordinates in the output array. See the link below for examples...
        # https://numpy.org/doc/stable/reference/generated/numpy.indices.html
        indices = np.indices((strided_result.shape[1], strided_result.shape[2]))

        # Make a copy of the argmax_arr because we will need two copies.
        # One for the x coordinates and one for the y coordinates.
        argmax_arr_x = argmax_arr.copy()
        # Floor divide by the stride's row size (number of columns) to get the x coordinate
        argmax_arr_x = argmax_arr_x // self.stride[1]
        # Modulus by the stride's row size (number of columns) to get the y coordinate. (uses the original argmax_arr for efficiency instead of another copy)
        argmax_arr_y = argmax_arr % self.stride[1]
        # Multiply each indice mask by the cooresponding row or column size
        indices_x = indices[0] * self.stride[0]
        indices_y = indices[1] * self.stride[1]
        # Loop through the depth of the input and apply the mask calculation to each input depth
        for depth in range(self.output_shape[0]):
            # Final calculation of the local_stride_index added with the (stride output coordinate * stride size)
            self.index_map_x[depth] = argmax_arr_x[depth] + indices_x
            self.index_map_y[depth] = argmax_arr_y[depth] + indices_y
            # Elementwise subtract top and left padding from the indices.
            # This is because we calculated the input index for the padded input and not the original input.
            self.index_map_x[depth] = self.index_map_x[depth] - self.padding[1][0]
            self.index_map_y[depth] = self.index_map_y[depth] - self.padding[2][0]
        return out

    def backward(self, output_gradient):
        input_gradient = np.zeros(self.input_shape, dtype="float64")
        for depth in range(output_gradient.shape[0]):
            for x in range(output_gradient.shape[1]):
                for y in range(output_gradient.shape[2]):
                    input_x_index = self.index_map_x[depth][x][y]
                    input_y_index = self.index_map_y[depth][x][y]
                    # Map the input index tracked through the translation mask to the corresponding output gradient
                    # Only the max values used in the forward pass will receive the gradient in the backward pass.
                    # Since only those max values contributed to the error.
                    # I do a += here because there could be multiple gradient values mapped to a single input
                    # if the stride is smaller than the pooling size.
                    input_gradient[depth][input_x_index][input_y_index] += output_gradient[depth][x][y]
        return input_gradient
    
    def parse_kernel(self, kernel):
        # checks for single integer
        if isinstance(kernel, int):
            return (kernel, kernel)
        # checks for single value in tuple
        elif isinstance(kernel, tuple) and len(kernel) == 1:
            return (kernel[0], kernel[0])
        elif isinstance(kernel, tuple) and len(kernel) == 2:
            # No changes necessary if tuple with 2 values.
            return kernel
        else:
            raise ValueError("Kernel size invalid. Please use one of the following formats...\n" \
                            "kernel_size = 3\n" \
                            "kernel_size = (3)\n" \
                            "kernel_size = (3,3)\n")

    # padding supports two padding formats depending on how you want to pad the various sides.
    # 2 or (2). Pads all sides
    # ((top, bottom), (left, right))
    # ((1,2),(2,1)). Pads sides individually
    def parse_padding(self, padding):
        # checks for single integer
        if isinstance(padding, int):
            # The extra tuple of zeros is for the depth dimension which needs no padding.
            return ((0, 0), (padding, padding), (padding, padding))
        # checks for single value in tuple
        elif isinstance(padding, tuple) and len(padding) == 1:
            return ((0, 0), (padding[0], padding[0]), (padding[0], padding[0]))
        elif isinstance(padding, tuple) and len(padding) == 2:
            # (height, width) or (3, 3)
            if isinstance(padding[0], int) and isinstance(padding[1], int):
                return ((0, 0), (padding[0], padding[0]), (padding[1], padding[1]))
            # ((top, bottom), (left, right))
            elif isinstance(padding[0], tuple) and isinstance(padding[1], tuple) and len(padding[0]) == 2 and len(padding[1]) == 2:
                return ((0, 0), (padding[0][0], padding[0][1]), (padding[1][0], padding[1][1]))
        else:
            raise ValueError("Padding size invalid. Please use one of the following formats...\n" \
                            "padding = 3\n" \
                            "padding = (3)\n" \
                            "padding = (3,3)\n" \
                            "padding = ((3,3), (3,3))\n")

    def sanity_checks(self):
        # Error checking to make sure the kernel and stride can evenly map across the input.
        dimension_1 = (self.input_shape[1] - self.kernel_size[0] + (self.padding[1][0] + self.padding[1][1])) / self.stride[0] + 1
        dimension_2 = (self.input_shape[2] - self.kernel_size[1] + (self.padding[2][0] + self.padding[2][1])) / self.stride[1] + 1
        if not dimension_1.is_integer() or not dimension_2.is_integer():
            raise ValueError("(input_size - kernel_size + (2 * padding)) / stride ... must be an integer.\n" \
                            "dimension 0 = depth\n" \
                            "dimension 1 = input height\n" \
                            "dimension 2 = input width\n" \
                            "This library does not handle cropping inputs for kernels / strides that do not evenly divide the input.")
        
        # If the padding is larger than the kernel size then you have values that will always be zero.
        # Because the output will only be calculated based on padding and not any input data
        if self.padding[1][0] >= self.kernel_size[0]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[1][0], self.kernel_size))
        elif self.padding[1][1] >= self.kernel_size[0]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[1][1], self.kernel_size))
        elif self.padding[2][0] >= self.kernel_size[1]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[2][0], self.kernel_size))
        elif self.padding[2][1] >= self.kernel_size[1]:
            raise ValueError("Padding size {} is greater than or equal to kernel size {}".format(self.padding[2][1], self.kernel_size))

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "(({}, {}, {}), kernel_size = {}, stride = {}, padding = {})".format(
            self.input_shape[0],
            self.input_shape[1],
            self.input_shape[2],
            self.kernel_size,
            self.stride,
            self.padding[1:]
        )


class Layer:
    def __init__(self):
        self.input = None
        self.output = None

    def forward(self, input):
        # TODO: return forward propagation output
        pass

    def backward(self, output_gradient):
        # TODO: update weights and return input gradient
        pass


import numpy as np
from scipy.ndimage import rotate, zoom, shift
import random
# For image debugging
from PIL import Image as im

# Helper Functions To Randomize Training Inputs
# These are all currently optimized for mnist

# Rotation method
def random_rotate_array(x, low_rotate_degrees = -20, high_rotate_degrees = 20):
    x = rotate(x, angle=random.randint(low_rotate_degrees, high_rotate_degrees), reshape=False)
    return x

# Translation method
def random_shift_array(x, vertical_shift_low = -3, vertical_shift_high = 3, horizontal_shift_low = -3, horizontal_shift_high = 3):
    x = shift(x, shift=(random.uniform(vertical_shift_low, vertical_shift_high),
                        random.uniform(horizontal_shift_low, horizontal_shift_high)))
    return x

# https://stackoverflow.com/questions/54633038/how-to-add-masking-noise-to-numpy-2-d-matrix-in-a-vectorized-manner
# Noise method
def random_noise_array(x, percentage_noise = 0.5, num_noise_iterations = 5, noise_value_low = 50, noise_value_high = 255):
    frac = percentage_noise / 100
    # Each iteration only applies a single noise value multiple times.
    # So typically you want a really low noise percentage and a higher number of iterations.
    # That way you get a bunch of unique noise values applied a small number of times.
    for i in range(num_noise_iterations):
        randomInt = random.randint(noise_value_low, noise_value_high)
        x[np.random.sample(size=x.shape) < frac] = randomInt
    return x

# Method from stackoverflow
# https://stackoverflow.com/questions/37119071/scipy-rotate-and-zoom-an-image-without-changing-its-dimensions
# Scaling method
# Was 0.75, 1.25
def random_clipped_zoom_array(img, zoom_factor=random.uniform(0.75, 1.4), **kwargs):

    h, w = img.shape[:2]

    # For multichannel images we don't want to apply the zoom factor to the RGB
    # dimension, so instead we create a tuple of zoom factors, one per array
    # dimension, with 1's for any trailing dimensions after the width and height.
    zoom_tuple = (zoom_factor,) * 2 + (1,) * (img.ndim - 2)

    # Zooming out
    if zoom_factor < 1:

        # Bounding box of the zoomed-out image within the output array
        zh = int(np.round(h * zoom_factor))
        zw = int(np.round(w * zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2

        # Zero-padding
        out = np.zeros_like(img)
        out[top:top+zh, left:left+zw] = zoom(img, zoom_tuple, **kwargs)

    # Zooming in
    elif zoom_factor > 1:

        # Bounding box of the zoomed-in region within the input array
        zh = int(np.round(h / zoom_factor))
        zw = int(np.round(w / zoom_factor))
        top = (h - zh) // 2
        left = (w - zw) // 2

        out = zoom(img[top:top+zh, left:left+zw], zoom_tuple, **kwargs)

        # `out` might still be slightly larger than `img` due to rounding, so
        # trim off any extra pixels at the edges
        trim_top = ((out.shape[0] - h) // 2)
        trim_left = ((out.shape[1] - w) // 2)
        out = out[trim_top:trim_top+h, trim_left:trim_left+w]

    # If zoom_factor == 1, just return the input array
    else:
        out = img
    return out

# Used to scale up numpy images for testing
def scale_image(np_array, scale_factor: int):
    return np.repeat(np.repeat(np_array, scale_factor, axis=0), scale_factor, axis=1)

# For debug only. Saves numpy array to file.
# You might have to multiple the numpy array by 255 if it was normalized
def save_image(np_array, file_name: str):
    # Create an image from the array
    data = im.fromarray(np_array)
    data = data.convert("L")
    
    # Saving the final output to file
    data.save(file_name)
    print("Saved Image... {}".format(file_name))


from nn.initializers.initializer import Initializer
from nn.optimizers.optimizer import Optimizer
import copy

class LayerProperties():

    def __init__(self, learning_rate = None, weight_initializer = None, bias_initializer = None, optimizer = None):
        self._learning_rate = learning_rate
        self._weight_initializer = weight_initializer
        self._bias_initializer = bias_initializer
        # Need independent optimizers because some optimizer require gradient shape.
        # And the gradient can be different for the weights and biases independently.
        self._weight_optimizer = copy.copy(optimizer)
        self._bias_optimizer = copy.copy(optimizer)
    
    @property
    def learning_rate(self) -> int:
        return self._learning_rate
    
    @learning_rate.setter
    def learning_rate(self, value):
        if value <= 0:
            raise ValueError("Learning rate must be greater than zero.")
        self._learning_rate = value

    @property
    def weight_initializer(self) -> Initializer:
        return self._weight_initializer
    
    @property
    def bias_initializer(self) -> Initializer:
        return self._bias_initializer

    @property
    def weight_optimizer(self) -> Optimizer:
        return self._weight_optimizer
    
    @property
    def bias_optimizer(self) -> Optimizer:
        return self._bias_optimizer
    
    #Need to add setters for the initializer and optimizer and check for the class types


from .optimizer import Optimizer
import numpy as np

class MomentumSGD(Optimizer):
    def __init__(self, mu=0.90):
        self.mu = mu
        self.v = None
    
    def calc(self, learning_rate, gradient):
        if self.v is None:
            self.v = np.zeros(gradient.shape)
        self.v = self.mu * self.v + learning_rate * gradient
        return -self.v


import numpy as np

def mean_squared_error(y_true, y_pred):
    return np.mean(np.power(y_true - y_pred, 2))

def mean_squared_error_prime(y_true, y_pred):
    return 2 * (y_pred - y_true) / np.size(y_true)


import numpy as np
from .initializer import Initializer

class Xavier(Initializer):
    def get(self, *shape):
        io = self.get_io_shape()
        input_neurons = np.prod(io[0])
        return np.random.randn(*shape) * np.sqrt(1 / input_neurons)


import numpy as np
import time
from nn.network.training_set import TrainingSet
from nn.layers.layer_properties import LayerProperties

from copy import deepcopy
from time import localtime, strftime

class Network():

    def __init__(
        self,
        layers,
        training_set: TrainingSet,
        loss,
        loss_prime,
        epochs = 1000,
        batch_size = 1,
        layer_properties: LayerProperties = None,
        data_augmentation = None,
        percent_error_threshold = None,
        verbose = True
    ) -> None:
        self.layers = layers
        self.training_set = training_set
        self.loss = loss
        self.loss_prime = loss_prime
        self.epochs = epochs
        self.batch_size = batch_size
        self.data_augmentation = data_augmentation
        self.percent_error_threshold = percent_error_threshold
        self.verbose = verbose
        # The total training time in minutes.
        self.total_training_time = 0
        # Per epoch errors. This can be used for debugging or visualization.
        self.per_epoch_errors = []

        # Optionally set the layer properties for all layers that utilize layer properties parameters
        if layer_properties is not None:
            self.layer_properties = layer_properties
            for layer in self.layers:
                if hasattr(layer, 'layer_properties'):
                    # Replace all layer defaults with any non "None" layer properties.
                    # This is just a lot of fancy code to allow you to override only 'some' of the default layer properties.
                    # Instead of forcing you to populate all the parameters every time.
                    for attr, _ in layer.layer_properties.__dict__.items():
                        if getattr(layer_properties, attr) is not None:
                            # copy is necessary to ensure that individual layer classes don't get shared instances of an optimizer
                            # optimizers such as momentum sgd require separate instances to track velocity
                            setattr(layer.layer_properties, attr, deepcopy(getattr(layer_properties, attr)))

    def predict(self, input):
        output = input
        for layer in self.layers:
            output = layer.forward(output)
        return output

    def train(self):
        # Reset start time and clear per epoch errors list
        start_time = time.time()
        self.per_epoch_errors.clear()
        
        if self.verbose:
            self.print_network_info()
            print("Start Date: {}".format(strftime("%Y-%m-%d", localtime())))
            print("Start Time: {}".format(strftime("%I:%M:%S %p", localtime())))
            print("Beginning training...")

        for epoch in range(self.epochs):

            #Optimization method            Samples in each gradient calculation        Weight updates per epoch
            #Batch Gradient Descent         The entire dataset	                        1
            #Minibatch Gradient Descent	    Consecutive subsets of the dataset	        n / size of minibatch
            #Stochastic Gradient Descent	Each sample of the dataset	                n
            #Increasing the batch size increases the number of epoches required for convergence
            for batch in self.iterate_minibatches(self.training_set.input_train, self.training_set.output_train, self.batch_size, shuffle=True):
                # Unpack batch training data
                input_train_batch, output_train_batch = batch
                # Track all gradients for the batch within a list
                # Using numpy empty over zeros for marginal performance
                gradients = np.empty(output_train_batch.shape)

                # Calculate the gradient for all training samples in the batch
                for sample_index, (input_train_sample, output_train_sample) in enumerate(zip(input_train_batch, output_train_batch)):
                    # Augment data while training if functions were provided
                    if self.data_augmentation is not None:
                        for func in self.data_augmentation:
                            input_train_sample = func(input_train_sample)

                    # Forward Propagation
                    prediction = self.predict(input_train_sample)

                    # Calculate Gradient
                    gradients[sample_index] = self.loss_prime(output_train_sample, prediction)
                    
                # Average all the gradients calculated in the batch
                gradient = np.mean(gradients, axis=0)

                # Backward Propagation
                for layer in reversed(self.layers):
                    gradient = layer.backward(gradient)

            # Calculate error statistics
            accuracy_train, accuracy_test = self.test()
            test_error_percentage = 1 - accuracy_test
            self.per_epoch_errors.append(test_error_percentage)

            if self.verbose:
                self.print_per_epoch_info(start_time, epoch, accuracy_train, accuracy_test)

            # If the error on the test set is less than the error threshold defined then we stop training.
            if self.percent_error_threshold is not None and test_error_percentage < self.percent_error_threshold:
                print("Error threshold {:.2f} has been reached.".format(self.percent_error_threshold))
                # Break out of the training loop
                break
        
        # Calculate total training time in minutes.
        end_time = time.time()
        time_elapsed_mins = (end_time - start_time) / 60
        self.total_training_time += time_elapsed_mins

        if self.verbose:
            print("Training Complete. Elapsed Time = {:.2f} seconds. Or {:.2f} minutes.".format(end_time - start_time, time_elapsed_mins))

    # Returns the accuracy against the training and test datasets
    def test(self):
        # Training Accuracy
        num_correct = 0
        num_incorrect = 0
        for input_train_sample, output_train_sample in zip(self.training_set.input_train, self.training_set.output_train):
            prediction = self.predict(input_train_sample)
            if self.training_set.post_processing(prediction) == self.training_set.post_processing(output_train_sample):
                num_correct += 1
            else:
                num_incorrect += 1
        accuracy_train = num_correct / (num_correct + num_incorrect)

        # Test Accuracy
        num_correct = 0
        num_incorrect = 0
        for input_train_sample, output_train_sample in zip(self.training_set.input_test, self.training_set.output_test):
            prediction = self.predict(input_train_sample)
            if self.training_set.post_processing(prediction) == self.training_set.post_processing(output_train_sample):
                num_correct += 1
            else:
                num_incorrect += 1
        accuracy_test = num_correct / (num_correct + num_incorrect)

        # Returns decimal format accuracy. i.e. 0.75 is 75%
        return accuracy_train, accuracy_test
    
    # Source: https://stackoverflow.com/questions/38157972/how-to-implement-mini-batch-gradient-descent-in-python
    # You should ideally shuffle the data. Take XOR for example if you have a batch size of 2.
    # And your batch pairs [0, 0] = [0] and [0, 1] = [1] it will average the gradient of these two examples every epoch.
    # Which means you will almost never reach a solution.
    def iterate_minibatches(self, inputs, targets, batchsize, shuffle=False):
        assert inputs.shape[0] == targets.shape[0]
        if shuffle:
            indices = np.arange(inputs.shape[0])
            np.random.shuffle(indices)
        for start_idx in range(0, inputs.shape[0], batchsize):
            end_idx = min(start_idx + batchsize, inputs.shape[0])
            if shuffle:
                excerpt = indices[start_idx:end_idx]
            else:
                excerpt = slice(start_idx, end_idx)
            yield inputs[excerpt], targets[excerpt]

    def print_network_info(self):

        print("===== Network Information =====")
        print("Network Architecture:")
        print("[")
        print(*self.layers, sep='\n')
        print("]\n")

        print("{:<15} {} {}".format("Training Data:", self.training_set.input_train_size, "samples"))
        print("{:<15} {} {}".format("Test Data:", self.training_set.input_test_size, "samples"))
        print("{:<15} {}".format("Loss Function:", self.loss.__name__))
        print("{:<15} {}".format("Epochs:", str(self.epochs)))

        if hasattr(self, 'layer_properties'):
            print("{:<15} {}".format("Learning Rate:", str(self.layer_properties.learning_rate)))
            print("{:<15} {}".format("Optimizer:", str(self.layer_properties._weight_optimizer.__class__.__name__)))

        print("{:<15} {}".format("Batch Size:", str(self.batch_size)))
        print("{:<15} {}".format("Verbose:", self.verbose))
        print("\n===== End Network Information =====\n")
    
    def print_per_epoch_info(self, start_time, epoch, accuracy_train, accuracy_test):
        # Calculate estimated training time remaining for my sanity
        end_time = time.time()
        time_elapsed_mins = (end_time - start_time) / 60
        time_per_epoch = time_elapsed_mins / (epoch + 1)
        epochs_remaining = self.epochs - (epoch + 1)
        training_time_remaining = time_per_epoch * epochs_remaining
        print("{}/{}, Accuracy Train = {:.2%}, Accuracy Test = {:.2%}, Time Remaining = {:.2f} minutes".format(
            (epoch+1), self.epochs, accuracy_train, accuracy_test, training_time_remaining))


from setuptools import setup, find_packages
# https://youtu.be/GaWs-LenLYE
# https://packaging.python.org/en/latest/tutorials/packaging-projects/
# Run 'pip install .' in the root to test the package locally

# https://towardsdatascience.com/how-to-upload-your-python-package-to-pypi-de1b363a1b3
# Create the dist directory
# python setup.py sdist

# Uploading to TestPyPi
# pip install twine
# twine upload --repository testpypi dist/*

# Specify index-url to download from TestPyPi
# pip install --extra-index-url https://test.pypi.org/simple/ aeronet-learn==1.0.1
# If you don't use the above command syntax then installing dependencies will fail.

# For uploading to prod PyPi. It will prompt for username / password
# twine upload dist/*

setup(
    name='aeronet-learn',
    version='1.0.4',
    author='Logan Zehm',
    author_email='logondz27@gmail.com',
    packages=find_packages(),
    url='https://github.com/Logon27/AeroNet',
    license='LICENSE',
    description='AeroNet is an educational neural network library. It is written entirely using python, numpy, and scipy.',
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    install_requires=[
        "numpy",
        "scipy",
        "dill",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
)


# The Base Network
from .network.network import Network
# Training Set
from .network.training_set import TrainingSet

# All Activation Functions
from .activations.leaky_relu import LeakyRelu
from .activations.relu import Relu
from .activations.sigmoid import Sigmoid
from .activations.softmax import Softmax
from .activations.tanh import Tanh

# All Standard Layers
from .layers.dense import Dense
from .layers.convolutional import Convolutional
from .layers.dropout import Dropout
from .layers.flatten import Flatten
from .layers.reshape import Reshape
from .layers.maxpooling2d import MaxPooling2D

# Layer Properties
from .layers.layer_properties import LayerProperties

# Initializers
from .initializers.normal import Normal
from .initializers.uniform import Uniform
from .initializers.zero import Zero
from .initializers.xavier import Xavier

# Optimizers
from .optimizers.sgd import SGD
from .optimizers.momentum_sgd import MomentumSGD

# Loss Functions
# Mean Squared Error
from .losses.mean_squared_error import mean_squared_error, mean_squared_error_prime
# Binary Cross Entropy
from .losses.binary_cross_entropy import binary_cross_entropy, binary_cross_entropy_prime
# Categorical Cross Entropy
from .losses.categorical_cross_entropy import categorical_cross_entropy, categorical_cross_entropy_prime

# Network File I/O
from .data_processing.file_io import save_network, load_network


import numpy as np
from nn.layers.layer import Layer

class Flatten(Layer):
    def __init__(self, input_shape):
        self.input_shape = input_shape

    def forward(self, input):
        # Needs to be a single column 2d array not a 1d array.
        return np.reshape(input, (input.size, 1))

    def backward(self, output_gradient):
        return np.reshape(output_gradient, self.input_shape)

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"


import numpy as np
from nn.layers.layer import Layer
from nn.layers.layer_properties import LayerProperties
from nn.initializers import *
from nn.optimizers import *
from copy import deepcopy

class Dense(Layer):

    def __init__(self, input_shape, output_shape, layer_properties: LayerProperties = None):
        # Default layer properties
        self.layer_properties = LayerProperties(learning_rate=0.05, weight_initializer=Uniform(), bias_initializer=Uniform(), optimizer=SGD())

        # Parse and update any optional layer properties
        self.parse_layer_properties(layer_properties)

        self.weights = self.layer_properties.weight_initializer.get(output_shape, input_shape)
        self.bias = self.layer_properties.weight_initializer.get(output_shape, 1)

    def forward(self, input):
        self.input = input
        return np.dot(self.weights, self.input) + self.bias

    def backward(self, output_gradient):
        weights_gradient = np.dot(output_gradient, self.input.T)
        input_gradient = np.dot(self.weights.T, output_gradient)
        self.weights += self.layer_properties.weight_optimizer.calc(self.layer_properties.learning_rate, weights_gradient)
        self.bias += self.layer_properties.bias_optimizer.calc(self.layer_properties.learning_rate, output_gradient)
        return input_gradient
    
    def parse_layer_properties(self, layer_properties):
        # Optionally set the layer properties for all layers that utilize layer properties parameters
        if layer_properties is not None:
            # Replace all layer defaults with any non "None" layer properties.
            # This is just a lot of fancy code to allow you to override only 'some' of the default layer properties.
            # Instead of forcing you to populate all the parameters every time.
            for attr, _ in layer_properties.__dict__.items():
                if getattr(layer_properties, attr) is not None:
                    # copy is necessary to ensure that individual layer classes don't get shared instances of an optimizer
                    # optimizers such as momentum sgd require separate instances to track velocity
                    setattr(self.layer_properties, attr, deepcopy(getattr(layer_properties, attr)))

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "({}, {})".format(self.weights.shape[1], self.weights.shape[0])


import numpy as np
from nn.layers.layer import Layer

class Dropout(Layer):
    def __init__(self, probability = 0.25):
        # Drop probability, if you want a success rate of 75% then you would
        # set a probability of 0.25 or 25% drop rate.
        self.probability = probability

    def forward(self, input):
        self.mask = np.random.binomial(1, self.probability, size=input.shape) / self.probability
        output = input * self.mask
        return output.reshape(input.shape)

    def backward(self, output_gradient):
        return output_gradient * self.mask

    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "({})".format(self.probability)


import numpy as np
from .initializer import Initializer

class Normal(Initializer):
    def __init__(self, mean=0, std=1):
        self.mean = mean
        self.std = std

    def get(self, *shape):
        return np.random.normal(self.mean, self.std, size=shape)


import dill
from nn import Network

#Provides the ability to save and load neural network objects from files.
#The entire network is saved as an object to a file.

def save_network(network: Network, file_path: str):
    with open(file_path, "wb") as file:
        print("Saving network to file {}...".format(file_path))
        dill.dump(network, file, dill.HIGHEST_PROTOCOL)
        print("Save complete.")

def load_network(file_path: str) -> Network:
    with open(file_path, "rb") as file:
        print("Loading network from file {}...".format(file_path))
        network = dill.load(file)
        print("Loading complete.")
        return network


import numpy as np
from nn.activations.activation import Activation

class Sigmoid(Activation):
    def __init__(self):
        def sigmoid(x):
            return 1 / (1 + np.exp(-x))

        def sigmoid_prime(x):
            s = sigmoid(x)
            return s * (1 - s)

        super().__init__(sigmoid, sigmoid_prime)
    
    # Modify string representation for network architecture printing
    def __str__(self):
        return self.__class__.__name__ + "()"